local save_file_is_cannon_unlocked, save_file_get_flags, save_file_set_flags, save_file_set_star_flags, network_send_to, network_local_index_from_global, obj_get_first_with_behavior_id, obj_set_model_extended, obj_get_next_with_same_behavior_id
    = save_file_is_cannon_unlocked, save_file_get_flags, save_file_set_flags, save_file_set_star_flags, network_send_to, network_local_index_from_global, obj_get_first_with_behavior_id, obj_set_model_extended, obj_get_next_with_same_behavior_id

gGlobalSyncTable.extraStars = 0
local file = get_current_save_file_num() - 1

local function set_extra_star_count()
    local count = 0
    for i = COURSE_NONE - 1, COURSE_END - 1, 1 do
        local unlocked = save_file_is_cannon_unlocked(file, i) ~= 0 -- ! Does not work with overworld
        if unlocked then
            count = count + 1
        end
        if i == COURSE_NONE - 1 and save_file_get_flags() & SAVE_FLAG_MOAT_DRAINED ~= 0 then
            count = count + 1
        end
    end
    gGlobalSyncTable.extraStars = count
end
set_extra_star_count()

local function save_file_set_cannon_flags_and_moat_drained(save_file, course)
    if course == COURSE_NONE - 1 then
        save_file_set_flags(SAVE_FLAG_MOAT_DRAINED)
    else
        save_file_set_star_flags(save_file, course, 1 << 7)
    end
end

---@param m MarioState
---@param obj Object
---@param intType integer
local function on_interact(m, obj, intType)
    if m.playerIndex ~= 0 then return end

    if intType == INTERACT_STAR_OR_KEY and obj_has_behavior_id(obj, id_bhvBowserKey) == 0 then
        if (obj.oBehParams >> 24) & 0xFF == 7 then
            if network_is_server() then
                save_file_set_cannon_flags_and_moat_drained(file, gNetworkPlayers[0].currCourseNum - 1)
            else
                network_send_to(network_local_index_from_global(0), true, {course = gNetworkPlayers[0].currCourseNum, stars = true})
            end
        end
    end
end

local function on_packet_recieve(datatable)
    if datatable.stars then
        save_file_set_cannon_flags_and_moat_drained(file, gNetworkPlayers[0].currCourseNum - 1)
        set_extra_star_count()
    end
end

---@param m MarioState
local function mario_update(m)
    if m.playerIndex ~= 0 then return end

    set_extra_star_count()
    local real_star_count = save_file_get_total_star_count(file, COURSE_MIN - 1, COURSE_END - 1)
    m.numStars = real_star_count + gGlobalSyncTable.extraStars
end

local spawned_stars = {
    --id_bhvHiddenStar, -- Spawn location for the secrets star
    id_bhvSpawnedStar, -- Spawned from boxes
    id_bhvSpawnedStarNoLevelExit, -- 100 coin stars
    --id_bhvHiddenRedCoinStar, -- Spawn location for the red coins star
    --id_bhvBowserCourseRedCoinStar, -- Spawn location for the red coins star
    id_bhvStarSpawnCoordinates, -- Spawned from enemies

    id_bhvStar -- Not spawned but needs to be in here
}

local function update()
    local course = gNetworkPlayers[0].currCourseNum - 1
    for _, starId in ipairs(spawned_stars) do
        local star = obj_get_first_with_behavior_id(starId)
        while star do
            if star.oBehParams >> 24 == 7 then
                if course ~= COURSE_NONE - 1 then
                    obj_set_model_extended(star, save_file_is_cannon_unlocked(file, course) ~= 0 and E_MODEL_TRANSPARENT_STAR or E_MODEL_STAR)
                else
                    obj_set_model_extended(star, save_file_get_flags() & SAVE_FLAG_MOAT_DRAINED ~= 0 and E_MODEL_TRANSPARENT_STAR or E_MODEL_STAR)
                end
            end
            star = obj_get_next_with_same_behavior_id(star)
        end
    end
end

hook_event(HOOK_ON_INTERACT, on_interact)
hook_event(HOOK_ON_PACKET_RECEIVE, on_packet_recieve)
hook_event(HOOK_MARIO_UPDATE, mario_update)
hook_event(HOOK_UPDATE, update)

_G.EighthStar = {
    exists = true,
}